@@ -3,7 +3,7 @@
  * 
  * This file is a part of NSIS.
  * 
- * Copyright (C) 1999-2009 Nullsoft and Contributors
+ * Copyright (C) 1999-2015 Nullsoft and Contributors
  * 
  * Licensed under the zlib/libpng license (the "License");
  * you may not use this file except in compliance with the License.
@@ -27,6 +27,31 @@
 #include "exec.h"
 #include "plugin.h"
 
+#ifndef LOAD_LIBRARY_SEARCH_USER_DIRS
+#define LOAD_LIBRARY_SEARCH_USER_DIRS 0x00000400
+#endif
+#ifndef LOAD_LIBRARY_SEARCH_SYSTEM32
+#define LOAD_LIBRARY_SEARCH_SYSTEM32 0x00000800
+#endif
+#ifndef SHTDN_REASON_FLAG_PLANNED
+#define SHTDN_REASON_FLAG_PLANNED 0x80000000
+#endif
+#ifndef SHTDN_REASON_MAJOR_APPLICATION
+#define SHTDN_REASON_MAJOR_APPLICATION 0x00040000
+#endif
+#ifndef SHTDN_REASON_MINOR_INSTALLATION
+#define SHTDN_REASON_MINOR_INSTALLATION 0x0002
+#endif
+#ifndef SHUTDOWN_RESTART
+#define SHUTDOWN_RESTART 0x00000004
+#endif
+#ifndef SHUTDOWN_FORCE_OTHERS
+#define SHUTDOWN_FORCE_OTHERS 0x00000001
+#endif
+#ifndef SHUTDOWN_GRACE_OVERRIDE
+#define SHUTDOWN_GRACE_OVERRIDE 0x00000020
+#endif
+
 #if !defined(NSIS_CONFIG_VISIBLE_SUPPORT) && !defined(NSIS_CONFIG_SILENT_SUPPORT)
 #error One of NSIS_CONFIG_SILENT_SUPPORT or NSIS_CONFIG_VISIBLE_SUPPORT must be defined.
 #endif
@@ -48,14 +73,14 @@ char *ValidateTempDir()
   if (!validpathspec(state_temp_dir))
     return NULL;
   addtrailingslash(state_temp_dir);
-  CreateDirectory(state_temp_dir, NULL);
+  CreateNormalDirectory(state_temp_dir);
   // state_language is used as a temp var here
   return my_GetTempFileName(state_language, state_temp_dir);
 }
 
 void *g_SHGetFolderPath;
 
-int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst,LPSTR lpszCmdParam, int nCmdShow)
+int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst, LPSTR lpszCmdParam, int nCmdShow)
 {
   int ret = 0;
   const char *m_Err = _LANG_ERRORWRITINGTEMP;
@@ -66,10 +91,45 @@ int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst,LPSTR lpszCmdParam,
   char seekchar=' ';
   char *cmdline;
 
-  InitCommonControls();
-
   SetErrorMode(SEM_NOOPENFILEERRORBOX | SEM_FAILCRITICALERRORS);
 
+  {
+    // bug #1125: Don't load modules from the application nor current directory.
+    // SetDefaultDllDirectories() allows us to restrict implicitly loaded and 
+    // dynamically loaded modules to just %windir%\System32 and directories 
+    // added with AddDllDirectory(). This prevents DLL search order attacks (CAPEC-471).
+    DWORD winver = GetVersion();
+    // CoCreateInstance(CLSID_ShellLink, ...) fails on Vista if SetDefaultDllDirectories is called
+    BOOL avoidwinbug = LOWORD(winver) == MAKEWORD(6, 0);
+    if (!avoidwinbug)
+    {
+      FARPROC fp = myGetProcAddress(MGA_SetDefaultDllDirectories);
+      if (fp) ((BOOL(WINAPI*)(DWORD))fp)(LOAD_LIBRARY_SEARCH_SYSTEM32|LOAD_LIBRARY_SEARCH_USER_DIRS);
+    }
+    LoadSystemLibrary("UXTHEME"); // On Vista OleInitialize calls NtUserCreateWindowEx and that pulls in UXTheme.dll
+    LoadSystemLibrary("USERENV"); // On Vista SHGetFileInfo ends up in SHELL32.kfapi::GetUserProfileDir and that pulls in UserEnv.dll
+    LoadSystemLibrary("SETUPAPI"); // On XP SHGetFileInfo ends up in CMountPoint::_InitLocalDriveHelper and that pulls in SetupAPI.dll
+  }
+
+  // Because myGetProcAddress now loads dlls with a full path 
+  // under GetSystemDirectory() the previous issues in <= v3.0b2 with 
+  // 'SetOutPath' and/or 'File "shfolder.dll"' no longer apply.
+  // All MGA dlls still need to be loaded early here because installers 
+  // running under WoW64 might disable WoW64 FS redirection in .onInit and 
+  // because GetSystemDirectory() can return the native system32 path we need
+  // the redirection to be turned off so LoadLibrary uses the correct folder.
+  // Note: We also import directly from KERNEL32, ADVAPI32 and SHELL32 so they 
+  // are exempt from this requirement and SHELL32 imports from SHLWAPI on 
+  // WoW64 systems and it is also on the KnownDLLs list so 
+  // SHLWAPI also gets a pass and that just leaves 
+#ifdef NSIS_SUPPORT_GETDLLVERSION
+  myGetProcAddress(MGA_GetFileVersionInfo); // VERSION
+#endif
+  g_SHGetFolderPath = myGetProcAddress(MGA_SHGetFolderPathA); // and SHFOLDER
+
+
+  InitCommonControls();
+
 #if defined(NSIS_SUPPORT_ACTIVEXREG) || defined(NSIS_SUPPORT_CREATESHORTCUT)
   {
     extern HRESULT g_hres;
@@ -77,18 +137,6 @@ int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst,LPSTR lpszCmdParam,
   }
 #endif
 
-  // load shfolder.dll before any script code is executed to avoid
-  // weird situations where SetOutPath or even the extraction of 
-  // shfolder.dll will cause unexpected behavior.
-  //
-  // this also prevents the following:
-  //
-  //  SetOutPath "C:\Program Files\NSIS" # maybe read from reg
-  //  File shfolder.dll
-  //  Delete $PROGRAMFILES\shfolder.dll # can't be deleted, as the
-  //                                    # new shfolder.dll is used
-  //                                    # to find its own path.
-  g_SHGetFolderPath = myGetProcAddress(MGA_SHGetFolderPathA);
 
   {
     // workaround for bug #1008632
@@ -214,16 +262,18 @@ int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst,LPSTR lpszCmdParam,
     }
     else
     {
-      int x;
+      int x, admin = UserIsAdminGrpMember();
 
-      mystrcat(state_temp_dir,"~nsu.tmp");
+      mystrcat(state_temp_dir,TEXT("~nsu"));
+      if (admin) mystrcat(state_temp_dir,TEXT("A")); // Don't lock down the directory used by non-admins
+      mystrcat(state_temp_dir,TEXT(".tmp"));
 
       // check if already running from uninstaller temp dir
       // this prevents recursive uninstaller calls
       if (!lstrcmpi(state_temp_dir,state_exe_directory))
         goto end;
 
-      CreateDirectory(state_temp_dir,NULL);
+      admin ? CreateRestrictedDirectory(state_temp_dir) : CreateNormalDirectory(state_temp_dir);
       SetCurrentDirectory(state_temp_dir);
 
       if (!state_install_directory[0])
@@ -297,9 +347,11 @@ end:
 #ifdef NSIS_SUPPORT_REBOOT
   if (g_exec_flags.reboot_called)
   {
+    const DWORD reason = SHTDN_REASON_FLAG_PLANNED | SHTDN_REASON_MAJOR_APPLICATION | SHTDN_REASON_MINOR_INSTALLATION;
     BOOL (WINAPI *OPT)(HANDLE, DWORD,PHANDLE);
     BOOL (WINAPI *LPV)(LPCTSTR,LPCTSTR,PLUID);
     BOOL (WINAPI *ATP)(HANDLE,BOOL,PTOKEN_PRIVILEGES,DWORD,PTOKEN_PRIVILEGES,PDWORD);
+    BOOL (WINAPI *IS)(LPTSTR,LPTSTR,DWORD,DWORD,DWORD);
     OPT=myGetProcAddress(MGA_OpenProcessToken);
     LPV=myGetProcAddress(MGA_LookupPrivilegeValueA);
     ATP=myGetProcAddress(MGA_AdjustTokenPrivileges);
@@ -316,7 +368,10 @@ end:
       }
     }
 
-    if (!ExitWindowsEx(EWX_REBOOT,0))
+    IS=myGetProcAddress(MGA_InitiateShutdown);
+    if ( (IS && !IS(NULL, NULL, 0, SHUTDOWN_RESTART | SHUTDOWN_FORCE_OTHERS | SHUTDOWN_GRACE_OVERRIDE, reason))
+      || (!ExitWindowsEx(EWX_REBOOT, reason))
+      )
       ExecuteCallbackFunction(CB_ONREBOOTFAILED);
   }
 #endif//NSIS_SUPPORT_REBOOT
