@@ -3,7 +3,7 @@
  * 
  * This file is a part of NSIS.
  * 
- * Copyright (C) 1999-2009 Nullsoft, Jeff Doozan and Contributors
+ * Copyright (C) 1999-2015 Nullsoft, Jeff Doozan and Contributors
  * 
  * Licensed under the zlib/libpng license (the "License");
  * you may not use this file except in compliance with the License.
@@ -382,9 +382,9 @@ FORCE_INLINE int NSISCALL ui_doinstall(void)
       static const char riched32[]="RichEd32";
       static const char richedit20a[]="RichEdit20A";
       static const char richedit[]="RichEdit";
-      if (!LoadLibrary(riched20))
+      if (!LoadSystemLibrary(riched20))
       {
-        LoadLibrary(riched32);
+        LoadSystemLibrary(riched32); // Win95 only ships with v1.0, NT4 has v2.0: web.archive.org/web/20030607222419/http://msdn.microsoft.com/library/en-us/shellcc/platform/commctls/richedit/richeditcontrols/aboutricheditcontrols.asp
       }
 
       // make richedit20a point to RICHEDIT
@@ -832,18 +832,41 @@ static BOOL CALLBACK UninstProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM l
 }
 #endif
 
+#ifndef _NSIS_NO_INT64_SHR
+#define NRT_U64Shr32(v,s) ( (v) >> (s) )
+#else
+#define NRT_U64Shr32 Int64ShrlMod32
+#endif
 
-static void NSISCALL SetSizeText(int dlgItem, int prefix, unsigned kb)
+static void NSISCALL SetSizeText64(int dlgItem, int prefix, ULARGE_INTEGER kb64)
 {
-  char scalestr[32], byte[32];
-  unsigned sh = 20;
+  CHAR scalestr[32], byte[32];
   int scale = LANG_GIGA;
+  UINT intgr, fract;
 
-  if (kb < 1024 * 1024) { sh = 10; scale = LANG_MEGA; }
-  if (kb < 1024) { sh = 0; scale = LANG_KILO; }
+  if (kb64.HighPart) // >= 4 TiB ?
+  {
+    kb64.QuadPart = NRT_U64Shr32(kb64.QuadPart, 20); // Convert from KiB to GiB
+    // wsprintf only supports the I64 size specifier on WinXP+.
+    // Older versions would crash because %s will use a bad pointer if we use "%I64u%s%s".
+    // Consequently we will only use the bottom 32-bits of the size (in GiB),
+    // this means we will display the wrong number if you have more than 4194303 TiB of free space.
+    intgr = kb64.LowPart;
+    fract = 0; // We don't even attempt to calculate this
+  }
+  else
+  {
+    unsigned sh = 20, kb = kb64.LowPart;
+    if (kb < 1024 * 1024) sh = 10, scale = LANG_MEGA;
+    if (kb < 1024) sh = 0, scale = LANG_KILO;
 
-  if (kb < (0xFFFFFFFF - ((1 << 20) / 20))) // check for overflow
-    kb += (1 << sh) / 20; // round numbers for better display (e.g. 1.59 => 1.6)
+    if (kb < (0xFFFFFFFF - ((1 << 20) / 20))) // check for overflow
+      kb += (1 << sh) / 20; // round numbers for better display (e.g. 1.59 => 1.6)
+
+     intgr = kb >> sh;
+     // 0x00FFFFFF mask is used to prevent overflow that causes bad results
+     fract = (((kb & 0x00FFFFFF) * 10) >> sh) % 10;
+  }
 
 #if _MSC_VER == 1200 // patch #1982084
   wsprintf(
@@ -853,17 +876,20 @@ static void NSISCALL SetSizeText(int dlgItem, int prefix, unsigned kb)
   wsprintf(
     g_tmp + mystrlen(g_tmp),
 #endif
-    "%u.%u%s%s",
-    kb >> sh,
-    (((kb & 0x00FFFFFF) * 10) >> sh) % 10, // 0x00FFFFFF mask is used to
-                                           // prevent overflow that causes
-                                           // bad results
+    TEXT("%u.%u%s%s"),
+    intgr, fract,
     GetNSISString(scalestr, scale),
     GetNSISString(byte, LANG_BYTE)
-  );
+    );
 
   my_SetDialogItemText(m_curwnd,dlgItem,g_tmp);
 }
+static void NSISCALL SetSizeText(int dlgItem, int prefix, unsigned kb)
+{
+  ULARGE_INTEGER kb64;
+  kb64.QuadPart = kb;
+  SetSizeText64(dlgItem, prefix, kb64);
+}
 
 static int NSISCALL _sumsecsfield(int idx)
 {
@@ -997,8 +1023,8 @@ static BOOL CALLBACK DirProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lPar
   {
     static char s[NSIS_MAX_STRLEN];
     int error = 0;
-    int available_set = 0;
-    unsigned total, available;
+    UINT total, available_set = FALSE;
+    ULARGE_INTEGER available;
 
     GetUIText(IDC_DIR,dir);
     if (!is_valid_instpath(dir))
@@ -1026,23 +1052,22 @@ static BOOL CALLBACK DirProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lPar
 
     // Test for and use the GetDiskFreeSpaceEx API
     {
-      BOOL (WINAPI *GDFSE)(LPCSTR, PULARGE_INTEGER, PULARGE_INTEGER, PULARGE_INTEGER) =
-          myGetProcAddress(MGA_GetDiskFreeSpaceExA);
+      BOOL (WINAPI *GDFSE)(LPCTSTR, PULARGE_INTEGER, PULARGE_INTEGER, PULARGE_INTEGER) =
+#ifdef _WIN64
+        GetDiskFreeSpaceEx;
+#else
+        myGetProcAddress(MGA_GetDiskFreeSpaceExA);
       if (GDFSE)
+#endif
       {
-        ULARGE_INTEGER available64;
         ULARGE_INTEGER a, b;
-        char *p;
+        CHAR *p;
         WORD *pw = NULL;
         while ((char *) pw != s) // trimslashtoend() cut the entire string
         {
-          if (GDFSE(s, &available64, &a, &b))
+          if (GDFSE(s, &available, &a, &b))
           {
-#ifndef _NSIS_NO_INT64_SHR
-            available = (int)(available64.QuadPart >> 10);
-#else
-            available = (int)(Int64ShrlMod32(available64.QuadPart, 10));
-#endif
+            available.QuadPart = NRT_U64Shr32(available.QuadPart, 10);
             available_set++;
             break;
           }
@@ -1057,37 +1082,37 @@ static BOOL CALLBACK DirProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lPar
         }
       }
     }
-
+#ifndef _WIN64
     if (!available_set)
     {
       DWORD spc, bps, fc, tc;
-      char *root;
+      CHAR *root;
 
       // GetDiskFreeSpaceEx accepts any path, but GetDiskFreeSpace accepts only the root
       mystrcpy(s,dir);
       root=skip_root(s);
-      if (root)
-        *root=0;
+      if (root) *root=0;
 
       // GetDiskFreeSpaceEx is not available
       if (GetDiskFreeSpace(s, &spc, &bps, &fc, &tc))
       {
-        available = (int)MulDiv(bps * spc, fc, 1 << 10);
+        available.QuadPart = (int)MulDiv(bps * spc, fc, 1 << 10);
         available_set++;
       }
     }
+#endif
+    total = (UINT) sumsecsfield(size_kb);
 
-    total = (unsigned) sumsecsfield(size_kb);
-
-    if (available_set && available < total)
-      error = NSIS_INSTDIR_NOT_ENOUGH_SPACE;
+    if (available_set)
+      if (available.QuadPart < total)
+        error = NSIS_INSTDIR_NOT_ENOUGH_SPACE;
 
     if (LANG_STR_TAB(LANG_SPACE_REQ)) {
       SetSizeText(IDC_SPACEREQUIRED,LANG_SPACE_REQ,total);
       if (available_set)
-        SetSizeText(IDC_SPACEAVAILABLE,LANG_SPACE_AVAIL,available);
+        SetSizeText64(IDC_SPACEAVAILABLE,LANG_SPACE_AVAIL,available);
       else
-        SetUITextNT(IDC_SPACEAVAILABLE,"");
+        SetUITextNT(IDC_SPACEAVAILABLE,TEXT(""));
     }
 
     g_exec_flags.instdir_error = error;
